library flutter_pattern_locker;

import 'dart:math';
import 'package:flutter/material.dart';

class PatternLocker extends StatefulWidget {
  final int rows; //行数
  final int columns; //列数
  final Size lockItemSize; //每个圆圈的大小
  final LockStyle style; //样式
  final bool isError; //是否显示错误
  final bool isKeepTrackOnComplete; //是否在完成绘制的时候保持轨迹
  final Function() onStart; //开始绘制
  final Function(String pwd) onComplete; //绘制结束
  PatternLocker(
      {this.rows = 3,
      this.columns = 3,
      this.lockItemSize = const Size(100, 100),
      this.isError = false,
      this.onStart,
      this.onComplete,
      this.isKeepTrackOnComplete = false,
      this.style,
      key})
      : super(key: key);

  @override
  _PatternLockerState createState() => _PatternLockerState();
}

class _PatternLockerState extends State<PatternLocker> {
  LockItemPoint markLockItemPoint;
  Offset currentPoint;
  List<LockItemPoint> lockItemCenterPoints = [];

  Widget buildColumn(int rowCount) {
    List<Widget> columns = [];
    for (int i = 0; i < widget.columns; i++) {
      LockItemPoint point;
      if (lockItemCenterPoints.length < widget.rows * widget.columns) {
        point = LockItemPoint(
            (i + widget.columns * rowCount + 1).toString(),
            Offset(
              widget.lockItemSize.width / 2 + widget.lockItemSize.width * i,
              widget.lockItemSize.height / 2 +
                  widget.lockItemSize.height * rowCount,
            ),
            widget.lockItemSize);
        lockItemCenterPoints.add(point);
      } else {
        point = lockItemCenterPoints[i + widget.columns * rowCount];
      }

      columns.add(_PartternLockItem(
        point: point,
        style: widget.style == null ? LockStyle() : widget.style,
        isError: widget.isError,
      ));
    }
    return Row(
        crossAxisAlignment: CrossAxisAlignment.center,
        mainAxisAlignment: MainAxisAlignment.center,
        children: columns);
  }

  List<Widget> buildRow() {
    List<Widget> rows = [];
    for (int i = 0; i < widget.rows; i++) {
      rows.add(buildColumn(i));
    }
    return rows;
  }

  void onTouch(Offset offset) {
    setState(() {
      currentPoint = offset;
    });
    for (var point in lockItemCenterPoints) {
      if (point.onTouch(offset)) {
        if (markLockItemPoint == null) {
          markLockItemPoint = point;
        } else {
          findLastLockItemPoint().next = point;
        }
        point.isMark = true;
        setState(() {
          point.isMark = true;
        });

        return;
      }
    }
  }

  LockItemPoint findLastLockItemPoint() {
    LockItemPoint point = markLockItemPoint;
    while (point.next != null) {
      point = point.next;
    }
    return point;
  }

  String getPassword() {
    String pwd = "";
    if (markLockItemPoint == null) return pwd;
    var point = markLockItemPoint;
    pwd = "$pwd${point.tag}";
    while (point.next != null) {
      point = point.next;
      pwd = "$pwd${point.tag}";
    }

    return pwd;
  }

  void resetAllPointAndLine() {
    markLockItemPoint = null;
    lockItemCenterPoints.forEach((point) {
      point.next = null;
      point.isMark = false;
    });
  }

  @override
  Widget build(BuildContext context) {
    return Container(
      alignment: Alignment.center,
      width: widget.lockItemSize.width * widget.columns,
      height: widget.lockItemSize.height * widget.rows,
      child: GestureDetector(
        onPanDown: (details) {
          resetAllPointAndLine();
          if (widget.onComplete != null) {
            widget.onStart();
          }
          onTouch(details.localPosition);
        },
        onPanUpdate: (details) {
          onTouch(details.localPosition);
        },
        onPanEnd: (details) {
          if (widget.onComplete != null) {
            widget.onComplete(getPassword());
          }
          if (!widget.isKeepTrackOnComplete) {
            resetAllPointAndLine();
          }
          setState(() {
            currentPoint = null;
          });
        },
        child: CustomPaint(
          child: Column(
            crossAxisAlignment: CrossAxisAlignment.center,
            mainAxisAlignment: MainAxisAlignment.center,
            children: buildRow(),
          ),
          painter: _PartternLockLinePainter(
              point: markLockItemPoint,
              currentPoint: currentPoint,
              style: widget.style == null ? LockStyle() : widget.style,
              isError: widget.isError),
        ),
      ),
    );
  }
}

class _PartternLockItem extends StatelessWidget {
  final LockItemPoint point;
  final LockStyle style;
  final bool isError;

  const _PartternLockItem({this.point, this.style, this.isError, Key key})
      : super(key: key);

  double _caculateAngle(LockItemPoint point) {
    // print(point);
    if (point.next == null) return 0;
    var angle = atan2(point.next.center.dx - point.center.dx,
        point.next.center.dy - point.center.dy);
    return pi - angle;
  }

  @override
  Widget build(BuildContext context) {
    return Container(
      width: point.size.width,
      height: point.size.height,
      child: Transform.rotate(
          angle: _caculateAngle(point),
          child: CustomPaint(
            painter: _PartternLockPainter(
              point,
              style: style,
              isError: isError,
            ),
          )),
    );
  }
}

class _PartternLockLinePainter extends CustomPainter {
  Paint _paint;
  LockItemPoint point;
  Offset currentPoint;
  LockStyle style;
  bool isError;
  _PartternLockLinePainter(
      {this.point, this.currentPoint, this.style, this.isError = false}) {
    _paint = Paint();

    _paint.strokeWidth = this.style.lineWidth;
  }
  @override
  void paint(Canvas canvas, Size size) {
    if (point == null) return;
    _paint.color = isError ? this.style.errorColor : this.style.selectedColor;
    Path path = Path();
    _paint.style = PaintingStyle.stroke;

    path.moveTo(point.center.dx, point.center.dy);
    while (point.next != null) {
      point = point.next;
      path.lineTo(point.center.dx, point.center.dy);
    }

    if (currentPoint != null) {
      path.lineTo(currentPoint.dx, currentPoint.dy);
    }
    canvas.drawPath(path, _paint);
  }

  @override
  bool shouldRepaint(CustomPainter oldDelegate) {
    return true;
  }
}

class _PartternLockPainter extends CustomPainter {
  Paint _paint;
  LockItemPoint point;
  LockStyle style;
  bool isError;

  _PartternLockPainter(this.point, {this.isError = false, this.style}) {
    // this.point = point;
    _paint = Paint();
    _paint.color = this.style.defaultColor;
    _paint.strokeWidth = this.style.borderWidth;
  }
  @override
  void paint(Canvas canvas, Size size) {
    _paint.color = this.point.isMark
        ? (this.isError ? this.style.errorColor : this.style.selectedColor)
        : this.style.defaultColor;

    var center = Offset(point.size.width / 2, point.size.height / 2);
    _paint.style = PaintingStyle.stroke;
    Path path = Path();
    Rect rect = Rect.fromCircle(center: center, radius: point.radius);
    path.addArc(rect, 0, 360);
    canvas.drawPath(path, _paint);

    if (!this.point.isMark) return;

    path = Path();
    rect = Rect.fromCircle(center: center, radius: point.radius);
    path.addArc(rect, 0, 360);
    _paint.color =
        (this.isError ? this.style.errorColor : this.style.selectedColor)
            .withAlpha(0x20);
    _paint.style = PaintingStyle.fill;
    canvas.drawPath(path, _paint);

    path = Path();
    rect = Rect.fromCircle(center: center, radius: point.radius / 3.5);
    path.addArc(rect, 0, 360);
    _paint.color =
        (this.isError ? this.style.errorColor : this.style.selectedColor);
    _paint.style = PaintingStyle.fill;
    canvas.drawPath(path, _paint);

    if (this.style.isShowArrow) {
      drawNextArrow(
          canvas, center, point.radius / 2.5, point.radius / 5, point.next);
    }
  }

  ///绘制箭头
  void drawNextArrow(Canvas canvas, Offset center, double arrowWidth,
      double arrowHeight, LockItemPoint nextPoint) {
    if (nextPoint == null) return;
    Path path = Path();
    _paint.style = PaintingStyle.fill;
    var startX = center.dx;
    var startY = center.dy - arrowHeight * 3;
    path.moveTo(startX, startY);
    path.lineTo(startX + arrowWidth / 2, startY + arrowHeight);
    path.lineTo(startX - arrowWidth / 2, startY + arrowHeight);
    path.lineTo(startX, startY);
    canvas.drawPath(path, _paint);
  }

  @override
  bool shouldRepaint(CustomPainter oldDelegate) {
    return true;
  }
}

class LockStyle {
  final double lineWidth;
  final double borderWidth;
  final Color selectedColor;
  final Color errorColor;
  final Color defaultColor;
  final bool isShowArrow;

  LockStyle({
    this.isShowArrow = true,
    this.lineWidth = 3.0,
    this.borderWidth = 1.0,
    this.selectedColor = Colors.blue,
    this.defaultColor = Colors.grey,
    this.errorColor = Colors.red,
  });
}

class LockItemPoint {
  String tag;
  Offset center;
  double radius;
  Size size;
  bool isMark = false;
  LockItemPoint next;

  LockItemPoint(this.tag, this.center, this.size) {
    radius = size.width * 0.4;
  }

  bool onTouch(Offset touch) {
    if (isMark) return false;
    var dx = (touch.dx - center.dx).abs();
    var dy = (touch.dy - center.dy).abs();
    var distance = sqrt(pow(dx, 2) + pow(dy, 2)); //获取touch点到圆心的距离
    if (distance <= radius) {
      return true;
    }
    return false;
  }
}
